@@ -1,18 +1,13 @@
 /*
- * Copyright (C) 1996-2017 The Squid Software Foundation and contributors
+ * DEBUG: section 16    Cache Manager API
  *
- * Squid software is distributed under GPLv2+ license and includes
- * contributions from numerous individuals and organizations.
- * Please see the COPYING and CONTRIBUTORS files for details.
  */
 
-/* DEBUG: section 16    Cache Manager API */
-
 #include "squid.h"
 #include "base/AsyncJobCalls.h"
 #include "base/TextException.h"
-#include "comm/Connection.h"
 #include "CommCalls.h"
+#include "comm/Connection.h"
 #include "errorpage.h"
 #include "globals.h"
 #include "HttpReply.h"
@@ -27,8 +22,8 @@ CBDATA_NAMESPACED_CLASS_INIT(Mgr, Forwarder);
 
 Mgr::Forwarder::Forwarder(const Comm::ConnectionPointer &aConn, const ActionParams &aParams,
                           HttpRequest* aRequest, StoreEntry* anEntry):
-    Ipc::Forwarder(new Request(KidIdentifier, 0, aConn, aParams), 10),
-    httpRequest(aRequest), entry(anEntry), conn(aConn)
+        Ipc::Forwarder(new Request(KidIdentifier, 0, aConn, aParams), 10),
+        httpRequest(aRequest), entry(anEntry), conn(aConn)
 {
     debugs(16, 5, HERE << conn);
     Must(Comm::IsConnOpen(conn));
@@ -36,7 +31,7 @@ Mgr::Forwarder::Forwarder(const Comm::ConnectionPointer &aConn, const ActionPara
     Must(entry != NULL);
 
     HTTPMSGLOCK(httpRequest);
-    entry->lock("Mgr::Forwarder");
+    entry->lock();
     EBIT_SET(entry->flags, ENTRY_FWD_HDR_WAIT);
 
     closer = asyncCall(16, 5, "Mgr::Forwarder::noteCommClosed",
@@ -52,7 +47,7 @@ Mgr::Forwarder::~Forwarder()
 
     HTTPMSGUNLOCK(httpRequest);
     entry->unregisterAbort();
-    entry->unlock("Mgr::Forwarder");
+    entry->unlock();
     cleanup();
 }
 
@@ -102,6 +97,17 @@ Mgr::Forwarder::noteCommClosed(const CommCloseCbParams& params)
     mustStop("commClosed");
 }
 
+/// called when Coordinator starts processing the request
+void
+Mgr::Forwarder::handleRemoteAck()
+{
+    Ipc::Forwarder::handleRemoteAck();
+
+    Must(entry != NULL);
+    EBIT_CLR(entry->flags, ENTRY_FWD_HDR_WAIT);
+    entry->complete();
+}
+
 /// send error page
 void
 Mgr::Forwarder::sendError(ErrorState *error)
@@ -119,4 +125,3 @@ Mgr::Forwarder::sendError(ErrorState *error)
     entry->flush();
     entry->complete();
 }
-
